
--[[ TODO
	
	Filter on update to improve search times
	Improve filter UI visual feedback
	Allow right click to remove filters
	
	Socket bonuses
	Stat weights
	
	flags = {
		Set, Resistances, Dodge, Parry, Charges, Azerite, Unique, Passives, Cooldown
	}
	
	Random tertiaries
	Random enchants
	
	----------------------------------------------------------------------------------------------------
	
	Data structure
	
	items = {
		ID, SubType, InventoryType, Rarity, Name, (Available), ItemLevel, Level, Bind,
		(Unique, Faction, Class), (Passives), Cooldown, DPS/Armor, Sockets,
		Strength, Agility, Intellect, Stamina, Critical Strike, Haste, Mastery, Versatility
	}
--]]

local bg

local function Chardev_Slash()
	-- Toggle UI visibility
	if bg then
		bg:SetShown(not bg:IsVisible())
		return
	end
	
	-- Localizations
	local floor, min, max, abs = math.floor, math.min, math.max, math.abs
	local create, resume, yield, status = coroutine.create, coroutine.resume, coroutine.yield, coroutine.status
	local remove, sort, concat = table.remove, table.sort, table.concat
	local tonumber, pairs, ipairs, collectgarbage, print = tonumber, pairs, ipairs, collectgarbage, print
	local lower, find, gmatch, gsub = string.lower, string.find, string.gmatch, string.gsub
	local ItemEventListener, GameTooltip, SetItemRef = ItemEventListener, GameTooltip, SetItemRef
	local GetItemInfoInstant, GetItemInfo, GetItemStats = GetItemInfoInstant, GetItemInfo, GetItemStats
	local GetTime, IsUsableItem, GetItemIcon = GetTime, IsUsableItem, GetItemIcon
	local C_AzeriteEmpoweredItem = C_AzeriteEmpoweredItem
	
	-- Data tables
	local itemTypes = {[2] = true, [4] = true}
	local inventoryTypes = {
		["INVTYPE_HEAD"] = 1, ["INVTYPE_NECK"] = 2, ["INVTYPE_SHOULDER"] = 3, ["INVTYPE_BODY"] = 4, ["INVTYPE_CHEST"] = 5,
		["INVTYPE_WAIST"] = 6, ["INVTYPE_LEGS"] = 7, ["INVTYPE_FEET"] = 8, ["INVTYPE_WRIST"] = 9, ["INVTYPE_HAND"] = 10,
		["INVTYPE_FINGER"] = 11, ["INVTYPE_TRINKET"] = 12, ["INVTYPE_WEAPON"] = 13, ["INVTYPE_SHIELD"] = 14,
		["INVTYPE_RANGED"] = 15, ["INVTYPE_CLOAK"] = 16, ["INVTYPE_2HWEAPON"] = 17, ["INVTYPE_ROBE"] = 20,
		["INVTYPE_WEAPONMAINHAND"] = 21, ["INVTYPE_WEAPONOFFHAND"] = 22, ["INVTYPE_HOLDABLE"] = 23, ["INVTYPE_RANGEDRIGHT"] = 26
	}
	
	local qualityColors = {
		[0] = "|cff9D9D9D", [1] = "|cffFFFFFF", [2] = "|cff1EFF00", [3] = "|cff0070DD", [4] = "|cffA335EE",
		[5] = "|cffFF8000", [6] = "|cffE6CC80", [7] = "|cff00CCFF", [8] = "|cff00CCFF"
	}
	
	local slots = {
		[1] = { -- One-Hand
			{{0, 13}, 	{0, 21}, 	{0, 22}},
			{{4, 13}, 	{4, 21}, 	{4, 22}},
			{{7, 13}, 	{7, 21}, 	{7, 22}},
			{{13, 13}, 	{13, 21}, 	{13, 22}},
			{{15, 13}, 	{15, 21}, 	{15, 22}},
			{{9, 13}, 	{9, 21}, 	{9, 22}},
			{{19, 26}}
		}, [2] = { -- Two-Hand
			{{1, 17}},
			{{5, 17}},
			{{8, 17}},
			{{6, 17}},
			{{10, 17}}
		}, [3] = { -- Ranged
			{{2, 15}},
			{{3, 26}},
			{{18, 26}}
		}, [4] = { -- Cloth
			{{101, 1}},
			{{101, 3}},
			{{101, 5}, 	{101, 20}},
			{{101, 9}},
			{{101, 10}},
			{{101, 6}},
			{{101, 7}},
			{{101, 8}},
		}, [5] = { -- Leather
			{{102, 1}},
			{{102, 3}},
			{{102, 5}, 	{102, 20}},
			{{102, 9}},
			{{102, 10}},
			{{102, 6}},
			{{102, 7}},
			{{102, 8}},
		}, [6] = { -- Mail
			{{103, 1}},
			{{103, 3}},
			{{103, 5}, 	{103, 20}},
			{{103, 9}},
			{{103, 10}},
			{{103, 6}},
			{{103, 7}},
			{{103, 8}},
		}, [7] = { -- Plate
			{{104, 1}},
			{{104, 3}},
			{{104, 5}, 	{104, 20}},
			{{104, 9}},
			{{104, 10}},
			{{104, 6}},
			{{104, 7}},
			{{104, 8}},
		}, [8] = { -- Miscellaneous
			{{106, 14}},
			{{100, 23}},
			{{100, 2}},
			{{101, 16}},
			{{100, 11}},
			{{100, 12}}
		}
	}
	
	local binds = {
		{0},
		{1, 4},
		{2}
	}
	
	local slotLabels1 = {"One-Hand", "Two-Hand", "Ranged", "Cloth", "Leather", "Mail", "Plate", "Miscellaneous"}
	local slotLabels2 = {
		{"Axe", "Mace", "Sword", "Fist Weapon", "Dagger", "Warglaive", "Wand"},
		{"Axe", "Mace", "Sword", "Polearm", "Staff"},
		{"Bow", "Crossbow", "Gun"},
		{"Head", "Shoulder", "Chest", "Wrist", "Hands", "Waist", "Legs", "Feet"},
		{"Head", "Shoulder", "Chest", "Wrist", "Hands", "Waist", "Legs", "Feet"},
		{"Head", "Shoulder", "Chest", "Wrist", "Hands", "Waist", "Legs", "Feet"},
		{"Head", "Shoulder", "Chest", "Wrist", "Hands", "Waist", "Legs", "Feet"},
		{"Shield", "Off-hand", "Neck", "Back", "Finger", "Trinket"}
	}
	
	local rarityLabels = {"Poor", "Common", "Uncommon", "Rare", "Epic", "Legendary", "Artifact", "Heirloom"}
	local bindLabels = {"None", "BoP", "BoE"}
	local filterLabels = {"iLvl", "Lvl", "D/Ar", "Sckt", "Str", "Agi", "Int", "Stam", "Crit", "Hste", "Mast", "Vers"}
	
	local IDs = Chardev_exclusions
	local exclusions = {}
	
	for a = 1, #IDs do
		exclusions[IDs[a]] = true
	end
	
	-- Unique variables
	local tasks = {}
	local items = {}
	local numItems = 0
	local results = {}
	local numResults = 0
	local filtered = {}
	local numFiltered = 0
	local arrange = 13
	local sorting = false
	local filters = {}
	local page = 1
	
	-- Throttling
	local check, message
	
	local function Process()
		if tasks[1] then
			if status(tasks[1]) == "dead" then
				remove(tasks, 1)
				
				if #tasks == 0 then
					if numResults ~= 1 then
						bg.title:SetText("Chardev | " .. numResults .. " items")
					else
						bg.title:SetText("Chardev | " .. numResults .. " item")
					end
				end
				
				collectgarbage()
				collectgarbage()
			else
				check, message = resume(tasks[1])
				
				if not check then
					bg.title:SetText("Chardev | Error")
					print("|cffF5DEB3Chardev|r | Error - try updating item list")
					print("|cffF5DEB3Chardev|r |", message)
				end
			end
		end
	end
	
	-- Update saved item list
	local updates
	local numUpdates
	local ID
	local found
	local instant
	local info
	local offset
	local stats
	local timestamp
	local remaining
	
	local function Update()
		bg.title:SetText("Chardev | Updating")
		
		updates = {}
		numUpdates = 0
		ID = 1
		remaining = {}
		
		repeat
			found = true
			
			for a = ID, ID + 999 do
				local instant = {GetItemInfoInstant(a)}
				
				if instant[1] then
					found = false
					
					-- Could filter some items by either subType or equipLocation, rather than both
					if itemTypes[instant[6]] and inventoryTypes[instant[4]] and not exclusions[a] then
						remaining[a] = ItemEventListener:AddCancelableCallback(a, function()
							timestamp = GetTime()
							info = {GetItemInfo(a)}
							
							offset = numUpdates * 24
							numUpdates = numUpdates + 1
							stats = GetItemStats(info[2])
							
							-- Record stats
							updates[offset + 1] = a
							
							if info[12] == 2 then
								updates[offset + 2] = info[13]
							else
								updates[offset + 2] = info[13] + 100
							end
							
							updates[offset + 3] = inventoryTypes[info[9]]
							updates[offset + 4] = info[3]
							updates[offset + 5] = info[1]
							updates[offset + 6] = 0
							updates[offset + 7] = info[4]
							updates[offset + 8] = info[5]
							updates[offset + 9] = info[14]
							updates[offset + 10] = 0
							updates[offset + 11] = 0
							updates[offset + 12] = 0
							updates[offset + 13] = 0
							updates[offset + 14] = IsUsableItem(a) and 1 or 0
							updates[offset + 15] = stats["ITEM_MOD_DAMAGE_PER_SECOND_SHORT"] or
														stats["RESISTANCE0_NAME"] or 0
							updates[offset + 16] = (stats["EMPTY_SOCKET_PRISMATIC"] or 0) +
														(stats["EMPTY_SOCKET_META"] or 0) +
														(stats["EMPTY_SOCKET_HYDRAULIC"] or 0) +
														(stats["EMPTY_SOCKET_COGWHEEL"] or 0)
							updates[offset + 17] = stats["ITEM_MOD_STRENGTH_SHORT"] or 0
							updates[offset + 18] = stats["ITEM_MOD_AGILITY_SHORT"] or 0
							updates[offset + 19] = stats["ITEM_MOD_INTELLECT_SHORT"] or 0
							updates[offset + 20] = stats["ITEM_MOD_STAMINA_SHORT"] or 0
							updates[offset + 21] = stats["ITEM_MOD_CRIT_RATING_SHORT"] or 0
							updates[offset + 22] = stats["ITEM_MOD_HASTE_RATING_SHORT"] or 0
							updates[offset + 23] = stats["ITEM_MOD_MASTERY_RATING_SHORT"] or 0
							updates[offset + 24] = stats["ITEM_MOD_VERSATILITY"] or 0
							
							remaining[a] = nil
						end)
					end
				end
			end
			
			ID = ID + 1000
			yield()
		until found
		
		-- Cancel any remaining queries after [1] second(s)
		while GetTime() - timestamp < 1 do
			yield()
		end
		
		for a, b in pairs(remaining) do
			b()
		end
		
		items = updates
		numItems = numUpdates
		
		-- Save items to disk
		Chardev_SavedVariable = concat(updates, "~") .. "~"
		bg.update:Enable()
	end
	
	-- Display searched results
	local result
	local offset
	local text
	
	local function Display()
		for a = page, page + 19 do
			result = "result" .. a - page + 1
			
			if numResults >= a then
				offset = (a - 1) * 10 + 1
				
				bg[result].icon:SetTexture(GetItemIcon(items[results[offset] + 1]))
				bg[result].link:SetText(qualityColors[items[results[offset] + 4]] .. items[results[offset] + 5])
				
				bg[result].stat1:SetText(items[results[offset] + 7])
				
				text = items[results[offset] + 8]
				if text == 0 then text = "-" end
				bg[result].stat2:SetText(text)
				
				text = floor(results[offset + 1] * 10 + 0.5) * 0.1
				if text == 0 then text = "-" end
				bg[result].stat3:SetText(text)
				
				text = items[results[offset] + 16]
				if text == 0 then text = "-" end
				bg[result].stat4:SetText(text)
				
				for b = 5, 12 do
					text = results[offset + b - 3]
					if text == 0 then text = "-" end
					bg[result]["stat" .. b]:SetText(text)
				end
				
				bg[result]:Show()
			else
				bg[result]:Hide()
			end
		end
	end
	
	-- Sort unsorted item list
	local order
	local filter
	local index
	local sorted
	
	local function Sort()
		bg.title:SetText("Chardev | Sorting")
		order = {}
		
		for a = 1, numFiltered do
			order[a] = a
		end
		
		-- Find current filter and sort
		filter = abs(arrange)
		yield()
		
		if filter == 13 then
			if arrange > 0 then
				sort(order, function(a, b)
						return items[filtered[(a - 1) * 10 + 1] + 5] < items[filtered[(b - 1) * 10 + 1] + 5] end)
			else
				sort(order, function(a, b)
						return items[filtered[(a - 1) * 10 + 1] + 5] > items[filtered[(b - 1) * 10 + 1] + 5] end)
			end
		elseif filter < 3 then
			index = filter + 6
			
			if arrange > 0 then
				sort(order, function(a, b)
						return items[filtered[(a - 1) * 10 + 1] + index] > items[filtered[(b - 1) * 10 + 1] + index] end)
			else
				sort(order, function(a, b)
						return items[filtered[(a - 1) * 10 + 1] + index] < items[filtered[(b - 1) * 10 + 1] + index] end)
			end
		elseif filter == 3 then
			if arrange > 0 then
				sort(order, function(a, b)
						return filtered[(a - 1) * 10 + 2] > filtered[(b - 1) * 10 + 2] end)
			else
				sort(order, function(a, b)
						return filtered[(a - 1) * 10 + 2] < filtered[(b - 1) * 10 + 2] end)
			end
		elseif filter == 4 then
			if arrange > 0 then
				sort(order, function(a, b)
						return items[filtered[(a - 1) * 10 + 1] + 16] > items[filtered[(b - 1) * 10 + 1] + 16] end)
			else
				sort(order, function(a, b)
						return items[filtered[(a - 1) * 10 + 1] + 16] < items[filtered[(b - 1) * 10 + 1] + 16] end)
			end
		else
			local index = filter - 2
			
			if arrange > 0 then
				sort(order, function(a, b)
						return filtered[(a - 1) * 10 + index] > filtered[(b - 1) * 10 + index] end)
			else
				sort(order, function(a, b)
						return filtered[(a - 1) * 10 + index] < filtered[(b - 1) * 10 + index] end)
			end
		end
		
		-- Assign sorted indices to values
		sorted = {}
		
		for a = 1, numFiltered, 1000 do
			yield()
			
			for b = a, min(numFiltered, a + 999) do
				for c = 1, 10 do
					sorted[(b - 1) * 10 + c] = filtered[(order[b] - 1) * 10 + c]
				end
			end
		end
		
		-- Show updated results
		results = sorted
		numResults = numFiltered
		
		sorting = false
		bg.scroll:SetMinMaxValues(1, max(1, numResults - 19))
		bg.scroll:SetValue(1)
		Display()
	end
	
	-- Filter, search and sort results
	local slotsDebounce
	local raritiesDebounce
	local bindsDebounce
	local debounce
	local up
	local down
	local weights
	
	local remaining
	local index2
	local stats
	local bonuses
	local delta
	local sum
	
	local function Search()
		bg.title:SetText("Chardev | Searching")
		
		-- Establish filters
		slotsDebounce = {}
		raritiesDebounce = {}
		bindsDebounce = {}
		debounce = true
		
		-- Slots
		for a = 0, 20 do
			slotsDebounce[a] = {}
		end
		
		for a = 100, 111 do
			slotsDebounce[a] = {}
		end
		
		for a = 1, #slots do
			for b = 1, #slots[a] do
				if bg.slot.container["dropDown" .. a].container["dropDown" .. b].selected then
					debounce = false
					
					for c = 1, #slots[a][b] do
						slotsDebounce[slots[a][b][c][1]][slots[a][b][c][2]] = true
					end
				end
			end
		end
		
		if debounce then
			for a = 1, #slots do
				for b = 1, #slots[a] do
					for c = 1, #slots[a][b] do
						slotsDebounce[slots[a][b][c][1]][slots[a][b][c][2]] = true
					end
				end
			end
		else
			debounce = true
		end
		
		-- Rarities
		for a = 1, 8 do
			if bg.rarity.container["dropDown" .. a].selected then
				debounce = false
				raritiesDebounce[a - 1] = true
			end
		end
		
		if debounce then
			for a = 1, 8 do
				raritiesDebounce[a - 1] = true
			end
		else
			debounce = true
		end
		
		-- Binds
		for a = 1, #binds do
			if bg.bind.container["dropDown" .. a].selected then
				debounce = false
				
				for b = 1, #binds[a] do
					bindsDebounce[binds[a][b]] = true
				end
			end
		end
		
		if debounce then
			for a = 1, #binds do
				for b = 1, #binds[a] do
					bindsDebounce[binds[a][b]] = true
				end
			end
		end
		
		-- Stat filters
		up = bg.scaleUp:GetChecked()
		down = bg.scaleDown:GetChecked()
		
		if up or down then
			filters[14] = tonumber(bg.ilvl:GetText())
		end
		
		filters[13] = lower(bg.name:GetText())
		
		if filters[13] == "name" then
			filters[13] = nil
		else
			filters[13] = gsub(filters[13], "[%(%)%.%+%-%*%?%[%]%^%$%%]", "%%%1")
		end
		
		for a = 1, 12 do
			filters[a] = tonumber(bg["filter" .. a]:GetText())
		end
		
		-- Stat weights
		weights = bg.weights:GetChecked()
		
		if not filters[3] then
			if weights then
				bg.filter3:SetText("£")
			else
				bg.filter3:SetText("D/Ar")
			end
		end
		
		-- Begin searching
		filtered = {}
		numFiltered = 0
		remaining = 0
		
		for a = 1, numItems, 1000 do
			for b = a, min(numItems, a + 999) do
				local index1 = (b - 1) * 24
				
				if slotsDebounce[items[index1 + 2]][items[index1 + 3]] and
						raritiesDebounce[items[index1 + 4]]											and
						bindsDebounce[items[index1 + 9]]											and
						(not filters[13] or find(lower(items[index1 + 5]), filters[13]))			and
						(not filters[1] or items[index1 + 7] >= filters[1])							and
						(not filters[2] or (filters[2] > 0 and items[index1 + 8] >= filters[2])		or
						(filters[2] == 0 and items[index1 + 8] == 0))								and
						(not filters[4] or (filters[4] > 0 and items[index1 + 16] >= filters[4])	or
						(filters[4] == 0 and items[index1 + 16] == 0) or weights)					then
						remaining = remaining + 1
					local itemString
					
					-- Optionally scale item level
					if filters[14] then
						delta = filters[14] - items[index1 + 7]
						
						if delta > 0 and up then
							bonuses = floor(abs(delta * 0.005) + 1.0005)
							itemString = {"item:", items[index1 + 1], "::::::::::::", bonuses, ":"}
							
							for c = 1, bonuses - 1 do
								itemString[c + 5] = "1672:"
							end
							
							itemString[#itemString + 1] = 1472 + delta - 200 * (bonuses - 1)
							itemString = concat(itemString)
						elseif delta < 0 and down then
							bonuses = floor(abs(delta * 0.01) + 1.005)
							itemString = {"item:", items[index1 + 1], "::::::::::::", bonuses, ":"}
							
							for c = 1, bonuses - 1 do
								itemString[c + 5] = "1372:"
							end
							
							itemString[#itemString + 1] = 1472 + delta + 100 * (bonuses - 1)
							itemString = concat(itemString)
						else
							itemString = "item:" .. items[index1 + 1]
						end
					else
						itemString = "item:" .. items[index1 + 1]
					end
					
					-- Load scaled item stats
					ItemEventListener:AddCallback(items[index1 + 1], function()
						remaining = remaining - 1
						index2 = numFiltered * 10
						stats = GetItemStats(itemString)
						
						filtered[index2 + 1] = index1
						filtered[index2 + 2] = stats["ITEM_MOD_DAMAGE_PER_SECOND_SHORT"] or stats["RESISTANCE0_NAME"] or 0
						filtered[index2 + 3] = stats["ITEM_MOD_STRENGTH_SHORT"] or 0
						filtered[index2 + 4] = stats["ITEM_MOD_AGILITY_SHORT"] or 0
						filtered[index2 + 5] = stats["ITEM_MOD_INTELLECT_SHORT"] or 0
						filtered[index2 + 6] = stats["ITEM_MOD_STAMINA_SHORT"] or 0
						filtered[index2 + 7] = stats["ITEM_MOD_CRIT_RATING_SHORT"] or 0
						filtered[index2 + 8] = stats["ITEM_MOD_HASTE_RATING_SHORT"] or 0
						filtered[index2 + 9] = stats["ITEM_MOD_MASTERY_RATING_SHORT"] or 0
						filtered[index2 + 10] = stats["ITEM_MOD_VERSATILITY"] or 0
						
						debounce = true
						
						if weights then
							-- Stat weighting
							sum = items[index1 + 16] * (filters[4] or 0)
							
							for c = 3, 10 do
								sum = sum + filtered[index2 + c] * (filters[c + 2] or 0)
							end
							
							filtered[index2 + 2] = sum
						else
							-- Stat filtering
							if filters[3] and (filtered[index2 + 2] < filters[3] or
									(filters[3] == 0 and filtered[index2 + 2] > 0))	then
								debounce = false
							else
								for c = 5, 12 do
									if filters[c] and (filtered[index2 + c - 2] < filters[c] or
											(filters[c] == 0 and filtered[index2 + c - 2] > 0))	then
										debounce = false
										break
									end
								end
							end
						end
						
						if debounce then
							numFiltered = numFiltered + 1
							
							--[[ [Test] Tooltip scanning
							local tooltip = C_TooltipInfo.GetHyperlink(itemString)
							TooltipUtil.SurfaceArgs(tooltip)
							
							for c, d in ipairs(tooltip.lines) do
								TooltipUtil.SurfaceArgs(d)
							end
							
							debounce = false
							
							for c = 2, #tooltip.lines do
								local text = tooltip.lines[c].leftText
								
								if text and text ~= " " then
									if find(text, "Equip:") or find(text, "Use:") or find(text, "hit:") then
										debounce = c
										break
									end
								end
							end
							
							if debounce then
								local numbers = {}
								
								for c = debounce, #tooltip.lines do
									local text = lower(tooltip.lines[c].leftText)
									
									if text and text ~= " " then
										if find(text, "damage") then break end
										if find(text, "heal") then break end
										if find(text, "mana") then break end
										if find(text, "set:") then break end
										
										if text and text ~= "" and text ~= " " then
											for number in gmatch(text, "%d+") do
												numbers[#numbers + 1] = tonumber(number)
											end
										end
									end
								end
								
								local greatest = 0
								
								for a = 1, #numbers do
									if numbers[a] > greatest then
										greatest = numbers[a]
									end
								end
								
								if greatest > 0 then
									filtered[index2 + 2] = greatest
									numFiltered = numFiltered + 1
								end
							end
							--]]
						end
					end)
				end
			end
			
			yield()
		end
		
		while remaining > 0 do
			yield()
		end
		
		Sort()
		bg.search:Enable()
	end
	
	-- One-time setup
	if not Chardev_SavedVariable then
		tasks[1] = create(function()
			bg.update:Disable()
			bg.title:SetText("Chardev | Updating")
			Update()
		end)
	end
	
	-- Unpack saved items
	tasks[#tasks + 1] = create(function()
		bg.title:SetText("Chardev | Loading")
		local conversion
		
		for a in gmatch(Chardev_SavedVariable, "([^~]+)~") do
			numItems = numItems + 1
			conversion = tonumber(a)
			
			if conversion then
				items[numItems] = conversion
			else
				items[numItems] = a
				
				if numItems % 24000 == 5 then
					yield()
				end
			end
		end
		
		numItems = floor(numItems / 24 + 0.5)
	end)
	
	-- Standardize frame creation
	local super = CreateFrame
	local function CreateFrame(frameType, frameName, parentFrame, inheritsFrame, dropDown1, dropDown2)
		local frame
		
		if inheritsFrame == "BackdropTemplate" then
			frame = super(frameType, frameName, parentFrame, inheritsFrame)
		else
			frame = super(frameType, frameName, parentFrame, "BackdropTemplate, " .. inheritsFrame)
		end
		
		frame:SetBackdrop({
			bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
			edgeFile = "Interface\\ChatFrame\\ChatFrameBackground",
			tile = false,
			tileSize = 0,
			edgeSize = 1,
			insets = {left = 0, right = 0, top = 0, bottom = 0}
		})
		
		if frameType == "Button" or frameType == "CheckButton" then
			frame:SetScript("OnEnter", function(self) self:SetBackdropColor(0.15, 0.15, 0.15) end)
			frame:SetScript("OnLeave", function(self) self:SetBackdropColor(0.2, 0.2, 0.2) end)
		end
		
		if frameType == "EditBox" then
			frame:SetFont("Interface\\Addons\\Chardev\\Assets\\Fonts\\Nunito\\Nunito-Regular.ttf", 16, "")
			frame:SetTextColor(0.961, 0.871, 0.702)
			frame:SetTextInsets(4, 4, 0, 0)
			frame:SetAutoFocus(false)
			frame:SetScript("OnEnter", function(self) self:SetBackdropColor(0.15, 0.15, 0.15) end)
			frame:SetScript("OnLeave", function(self) self:SetBackdropColor(0.2, 0.2, 0.2) end)
			frame:SetScript("OnEscapePressed", function(self) self:ClearFocus() end)
		end
		
		if parentFrame == UIParent then
			frame:SetFrameStrata("DIALOG")
			frame:SetFrameLevel(10)
			frame:SetMovable(true)
			frame:SetBackdropColor(0.1, 0.1, 0.1)
			frame:SetBackdropBorderColor(0.05, 0.05, 0.05)
			frame:SetScript("OnMouseDown", function(self) self:StartMoving() end)
			frame:SetScript("OnMouseUp", function(self) self:StopMovingOrSizing() end)
			
			local super = frame.CreateFontString
			frame.CreateFontString = function(...)
				local text = super(...)
				text:SetFont("Interface\\Addons\\Chardev\\Assets\\Fonts\\Nunito\\Nunito-Regular.ttf", 22, "")
				text:SetTextColor(0.961, 0.871, 0.702)
				
				return text
			end
		else
			frame:SetBackdropColor(0.2, 0.2, 0.2)
			frame:SetBackdropBorderColor(0.2, 0.2, 0.2)
			
			local super = frame.CreateFontString
			frame.CreateFontString = function(...)
				local text = super(...)
				text:SetFont("Interface\\Addons\\Chardev\\Assets\\Fonts\\Nunito\\Nunito-Regular.ttf", 16, "")
				text:SetTextColor(0.961, 0.871, 0.702)
				
				return text
			end
		end
		
		if inheritsFrame == "OptionsSliderTemplate" then
			frame.High:SetText("")
			frame.Low:SetText("")
			frame.NineSlice:Hide()
			
			frame.Thumb:SetColorTexture(0.3, 0.3, 0.3)
			frame.Thumb:SetSize(24, 24)
		end
		
		if dropDown1 then
			frame:SetSize(116, 24)
			frame.text = frame:CreateFontString()
			frame.text:SetPoint("CENTER")
			frame.container = super("Frame", nil, bg)
			frame.container:Hide()
			frame:SetScript("OnEnter", function(self)
				self:SetBackdropColor(0.15, 0.15, 0.15)
			end)
			
			frame:SetScript("OnLeave", function(self)
				self:SetBackdropColor(0.2, 0.2, 0.2)
			end)
			
			frame:SetScript("OnMouseDown", function(self) self.text:SetPoint("CENTER", 2, -2) end)
			frame:SetScript("OnMouseUp", function(self, button)
				self.text:SetPoint("CENTER")
				
				if button == "LeftButton" and self:IsMouseOver() then
					if self.container:IsVisible() then
						self.container:Hide()
					else
						bg.slot.container:Hide()
						bg.rarity.container:Hide()
						bg.bind.container:Hide()
						self.container:Show()
					end
				end
			end)
			
			for a = 1, #dropDown1 do
				local index = "dropDown" .. a
				
				frame.container[index] = CreateFrame("Button", nil, frame.container, "BackdropTemplate")
				frame.container[index]:SetPoint("TOP", frame, "TOP", 0, -24 * a)
				frame.container[index]:SetSize(116, 24)
				frame.container[index]:SetBackdropColor(0.3, 0.3, 0.3)
				frame.container[index]:SetBackdropBorderColor(0.25, 0.25, 0.25)
				frame.container[index].selected = false
				frame.container[index].text = frame.container[index]:CreateFontString()
				frame.container[index].text:SetPoint("LEFT", 4, 0)
				frame.container[index].text:SetText(dropDown1[a])
				
				if not dropDown2 then
					frame.container[index]:SetScript("OnEnter", function(self)
						if not self.selected then
							self:SetBackdropColor(0.2, 0.2, 0.2)
						end
					end)
					
					frame.container[index]:SetScript("OnLeave", function(self)
						if not self.selected then
							self:SetBackdropColor(0.3, 0.3, 0.3)
						end
					end)
					
					frame.container[index]:SetScript("OnMouseDown", function(self) self.text:SetPoint("LEFT", 6, -2) end)
					frame.container[index]:SetScript("OnMouseUp", function(self, button)
						self.text:SetPoint("LEFT", 4, 0)
						
						if button == "LeftButton" and self:IsMouseOver() then
							self.selected = not self.selected
						end
					end)
				else
					frame.container[index]:SetScript("OnEnter", function(self)
						self.container:Show()
						
						if not self.selected then
							self:SetBackdropColor(0.2, 0.2, 0.2)
						end
					end)
					
					frame.container[index]:SetScript("OnLeave", function(self)
						if not self.container.dropDown1:IsMouseOver(1, -1, -1, 1) then
							self.container:Hide()
						end
						
						if not self.selected then
							self:SetBackdropColor(0.3, 0.3, 0.3)
						end
					end)
					
					frame.container[index]:SetScript("OnMouseDown", function(self) self.text:SetPoint("LEFT", 6, -2) end)
					frame.container[index]:SetScript("OnMouseUp", function(self, button)
						self.text:SetPoint("LEFT", 4, 0)
						
						if button == "LeftButton" and self:IsMouseOver() then
							self.selected = not self.selected
							
							for b = 1, #dropDown2[a] do
								index2 = "dropDown" .. b
								self.container[index2].selected = self.selected
								
								if self.selected then
									self.container[index2]:SetBackdropColor(0.2, 0.2, 0.2)
								else
									self.container[index2]:SetBackdropColor(0.3, 0.3, 0.3)
								end
							end
						end
					end)
					
					frame.container[index].container = super("Frame", nil, frame.container[index])
					frame.container[index].container:Hide()
					
					local index2
					
					for b = 1, #dropDown2[a] do
						index2 = "dropDown" .. b
						
						frame.container[index].container[index2] =
								CreateFrame("Button", nil, frame.container[index].container, "BackdropTemplate")
						frame.container[index].container[index2]:SetPoint(
								"LEFT", frame.container[index], "RIGHT", 0, -24 * (b - 1))
						frame.container[index].container[index2]:SetSize(116, 24)
						frame.container[index].container[index2]:SetBackdropColor(0.3, 0.3, 0.3)
						frame.container[index].container[index2]:SetBackdropBorderColor(0.25, 0.25, 0.25)
						frame.container[index].container[index2].selected = false
						frame.container[index].container[index2].text =
								frame.container[index].container[index2]:CreateFontString()
						frame.container[index].container[index2].text:SetPoint("LEFT", 4, 0)
						frame.container[index].container[index2].text:SetText(dropDown2[a][b])
						frame.container[index].container[index2]:EnableMouse(true)
						frame.container[index].container[index2]:SetScript("OnEnter", function(self)
							if not self.selected then
								self:SetBackdropColor(0.2, 0.2, 0.2)
							end
						end)
						
						frame.container[index].container[index2]:SetScript("OnLeave", function(self)
							local debounce = true
							
							for c = 1, #dropDown2[a] do
								if c ~= b and
										frame.container[index].container["dropDown" .. c]:IsMouseOver(1, -1, -1, 1)	then
									debounce = false
									break
								end
							end
							
							if debounce then
								frame.container[index].container:Hide()
							end
							
							if not self.selected then
								self:SetBackdropColor(0.3, 0.3, 0.3)
							end
						end)
						
						frame.container[index].container[index2]:SetScript(
								"OnMouseDown", function(self) self.text:SetPoint("LEFT", 6, -2) end)
						frame.container[index].container[index2]:SetScript("OnMouseUp", function(self, button)
							self.text:SetPoint("LEFT", 4, 0)
							
							if button == "LeftButton" and self:IsMouseOver() then
								self.selected = not self.selected
								
								local debounce = false
								
								for c = 1, #dropDown2[a] do
									if frame.container[index].container["dropDown" .. c].selected then
										debounce = true
										break
									end
								end
								
								frame.container[index].selected = debounce
								
								if debounce then
									frame.container[index]:SetBackdropColor(0.2, 0.2, 0.2)
								else
									frame.container[index]:SetBackdropColor(0.3, 0.3, 0.3)
								end
							end
						end)
					end
				end
			end
		end
		
		return frame
	end
	
	-- Create UI
	bg = CreateFrame("Frame", nil, UIParent, "BackdropTemplate")
	bg:SetPoint("CENTER")
	bg:SetSize(1048, 640)
	bg:SetPropagateKeyboardInput(true)
	bg.title = bg:CreateFontString()
	bg.title:SetPoint("TOPLEFT", 14, -14)
	bg:SetScript("OnMouseWheel", function(_, direction) bg.scroll:SetValue(page - direction) end)
	bg:SetScript("OnKeyDown", function(self, key)
		if key == "ESCAPE" then
			self:Hide()
		end
	end)
	
	bg:SetScript("OnUpdate", Process)
	
	-- Top row
	bg.ilvl = CreateFrame("EditBox", nil, bg, "BackdropTemplate")
	bg.ilvl:SetPoint("TOPRIGHT", -606, -14)
	bg.ilvl:SetSize(84, 24)
	bg.ilvl:SetText("Scale iLvl")
	bg.ilvl:SetScript("OnEditFocusGained", function(self)
		if not self:IsNumeric() then
			self:SetText("")
			self:SetTextColor(0.95, 0.95, 0.95)
			self:SetNumeric(true)
			self:SetMaxLetters(3)
		end
	end)
	
	bg.ilvl:SetScript("OnEditFocusLost", function(self)
		self:HighlightText(0, 0)
		
		if self:GetText() == "" then
			self:SetNumeric(false)
			self:SetMaxLetters(255)
			self:SetTextColor(0.961, 0.871, 0.702)
			self:SetText("Scale iLvl")
		end
	end)
	
	bg.ilvl:SetScript("OnEnterPressed", function(self)
		self:ClearFocus()
		bg.search:Click()
	end)
	
	bg.scaleDown = CreateFrame("CheckButton", nil, bg, "BackdropTemplate")
	bg.scaleDown:SetPoint("TOPRIGHT", -694, -14)
	bg.scaleDown:SetSize(24, 24)
	bg.scaleDown:SetChecked(true)
	bg.scaleDown.text = bg.scaleDown:CreateFontString()
	bg.scaleDown.text:SetPoint("CENTER", 0, -1)
	bg.scaleDown.text:SetText("V")
	bg.scaleDown.texture = bg.scaleDown:CreateTexture()
	bg.scaleDown.texture:SetColorTexture(0.3, 0.3, 0.3)
	bg.scaleDown.texture:SetPoint("CENTER")
	bg.scaleDown.texture:SetSize(18, 18)
	bg.scaleDown:SetCheckedTexture(bg.scaleDown.texture)
	bg.scaleDown:SetScript("OnMouseDown", function(self)
		self.text:SetPoint("CENTER", 2, -3)
	end)
	
	bg.scaleDown:SetScript("OnMouseUp", function(self)
		self.text:SetPoint("CENTER", 0, -1)
	end)
	
	bg.scaleUp = CreateFrame("CheckButton", nil, bg, "BackdropTemplate")
	bg.scaleUp:SetPoint("TOPRIGHT", -578, -14)
	bg.scaleUp:SetSize(24, 24)
	bg.scaleUp:SetChecked(true)
	bg.scaleUp.text = bg.scaleUp:CreateFontString()
	bg.scaleUp.text:SetPoint("CENTER", 0, -4)
	bg.scaleUp.text:SetFont("Interface\\Addons\\Chardev\\Assets\\Fonts\\Nunito\\Nunito-Regular.ttf", 22, "")
	bg.scaleUp.text:SetText("^")
	bg.scaleUp.texture = bg.scaleUp:CreateTexture()
	bg.scaleUp.texture:SetColorTexture(0.3, 0.3, 0.3)
	bg.scaleUp.texture:SetPoint("CENTER")
	bg.scaleUp.texture:SetSize(18, 18)
	bg.scaleUp:SetCheckedTexture(bg.scaleUp.texture)
	bg.scaleUp:SetScript("OnMouseDown", function(self)
		self.text:SetPoint("CENTER", 2, -6)
	end)
	
	bg.scaleUp:SetScript("OnMouseUp", function(self)
		self.text:SetPoint("CENTER", 0, -4)
	end)
	
	bg.slot = CreateFrame("Button", nil, bg, "BackdropTemplate", slotLabels1, slotLabels2)
	bg.slot:SetPoint("TOPRIGHT", -458, -14)
	bg.slot.text:SetText("Gear Type")
	
	bg.rarity = CreateFrame("Button", nil, bg, "BackdropTemplate", rarityLabels)
	bg.rarity:SetPoint("TOPRIGHT", -338, -14)
	bg.rarity.text:SetText("Rarity")
	
	local dropDown
	
	for a = 1, 8 do
		dropDown = "dropDown" .. a
		bg.rarity.container[dropDown].text:SetText(qualityColors[a - 1] .. bg.rarity.container[dropDown].text:GetText())
	end
	
	bg.bind = CreateFrame("Button", nil, bg, "BackdropTemplate", bindLabels)
	bg.bind:SetPoint("TOPRIGHT", -218, -14)
	bg.bind.text:SetText("Bind Type")
	
	bg.search = CreateFrame("Button", nil, bg, "BackdropTemplate")
	bg.search:SetPoint("TOPRIGHT", -130, -14)
	bg.search:SetSize(84, 24)
	bg.search.text = bg.search:CreateFontString()
	bg.search.text:SetPoint("CENTER")
	bg.search.text:SetText("Search")
	bg.search:SetScript("OnMouseDown", function(self)
		if self:IsEnabled() then
			self.text:SetPoint("CENTER", 2, -2)
		end
	end)
	
	bg.search:SetScript("OnMouseUp", function(self, button)
		if self:IsEnabled() then
			self.text:SetPoint("CENTER")
			
			if button == "LeftButton" and self:IsMouseOver() then
				self:Click()
			end
		end
	end)
	
	bg.search:SetScript("OnClick", function(self)
		if self:IsEnabled() then
			self:Disable()
			bg.slot.container:Hide()
			bg.rarity.container:Hide()
			bg.bind.container:Hide()
			
			tasks[#tasks + 1] = create(Search)
		end
	end)
	
	bg.update = CreateFrame("Button", nil, bg, "BackdropTemplate")
	bg.update:SetPoint("TOPRIGHT", -42, -14)
	bg.update:SetSize(84, 24)
	bg.update.text = bg.update:CreateFontString()
	bg.update.text:SetPoint("CENTER")
	bg.update.text:SetText("Update")
	bg.update:SetScript("OnMouseDown", function(self)
		if self:IsEnabled() then
			self.text:SetPoint("CENTER", 2, -2)
		end
	end)
	
	bg.update:SetScript("OnMouseUp", function(self, button)
		if self:IsEnabled() then
			self.text:SetPoint("CENTER")
			
			if button == "LeftButton" and self:IsMouseOver() then
				self:Disable()
				tasks[#tasks + 1] = create(Update)
			end
		end
	end)
	
	bg.close = CreateFrame("Button", nil, bg, "BackdropTemplate")
	bg.close:SetPoint("TOPRIGHT", -14, -14)
	bg.close:SetSize(24, 24)
	bg.close.text = bg.close:CreateFontString()
	bg.close.text:SetPoint("CENTER")
	bg.close.text:SetFont("Interface\\Addons\\Chardev\\Assets\\Fonts\\Nunito\\Nunito-Regular.ttf", 20, "")
	bg.close.text:SetText("x")
	bg.close:SetScript("OnMouseDown", function(self) self.text:SetPoint("CENTER", 2, -2) end)
	bg.close:SetScript("OnMouseUp", function(self, button)
		self.text:SetPoint("CENTER")
		
		if button == "LeftButton" and self:IsMouseOver() then
			bg:Hide()
		end
	end)
	
	-- Sorting
	bg.name = CreateFrame("EditBox", nil, bg, "BackdropTemplate")
	bg.name:SetPoint("TOPLEFT", 14, -42)
	bg.name:SetSize(344, 24)
	bg.name:SetBackdropColor(0.15, 0.15, 0.15)
	bg.name:SetText("Name")
	bg.name:SetScript("OnMouseDown", function(self, button)
		if button == "LeftButton" then
			self:SetTextInsets(6, 4, 0, -2)
		end
	end)
	
	bg.name:SetScript("OnMouseUp", function(self, button)
		if button == "LeftButton" then
			self:SetTextInsets(4, 4, 0, 0)
			
			if self:IsMouseOver() and not sorting then
				if abs(arrange) == 13 then
					arrange = arrange * -1
				else
					bg["filter" .. abs(arrange)]:SetBackdropColor(0.2, 0.2, 0.2)
					arrange = 13
				end
				
				self:ClearFocus()
				sorting = true
				tasks[#tasks + 1] = create(Sort)
			end
		end
	end)
	
	bg.name:SetScript("OnEditFocusGained", function(self)
		if self:GetText() == "Name" then
			self:SetText("")
			self:SetTextColor(0.95, 0.95, 0.95)
		end
		
		if self:GetTextInsets() == 6 then
			self:ClearFocus()
		end
	end)
	
	bg.name:SetScript("OnEditFocusLost", function(self)
		self:HighlightText(0, 0)
		
		if self:GetText() == "" then
			self:SetTextColor(0.961, 0.871, 0.702)
			self:SetText("Name")
		end
	end)
	
	bg.name:SetScript("OnEnterPressed", function(self)
		self:ClearFocus()
		bg.search:Click()
	end)
	
	bg.name:SetScript("OnEnter", function(self)
		if abs(arrange) ~= 13 then
			self:SetBackdropColor(0.15, 0.15, 0.15)
		end
	end)
	
	bg.name:SetScript("OnLeave", function(self)
		if abs(arrange) ~= 13 then
			self:SetBackdropColor(0.2, 0.2, 0.2)
		end
	end)
	
	local filter
	
	for a = 1, 12 do
		filter = "filter" .. a
		
		bg[filter] = CreateFrame("EditBox", nil, bg, "BackdropTemplate")
		bg[filter]:SetPoint("TOPLEFT", 304 + a * 54, -42)
		bg[filter]:SetSize(54, 24)
		bg[filter]:SetText(filterLabels[a])
		bg[filter]:SetMaxLetters(4)
		bg[filter].arrange = 0
		bg[filter]:SetScript("OnMouseDown", function(self, button)
			if button == "LeftButton" then
				self:SetTextInsets(6, 4, 0, -2)
			end
		end)
		
		bg[filter]:SetScript("OnMouseUp", function(self, button)
			if button == "LeftButton" then
				self:SetTextInsets(4, 4, 0, 0)
				
				if self:IsMouseOver() and not sorting then
					if abs(arrange) == a then
						arrange = arrange * -1
					else
						if abs(arrange) < 13 then
							bg["filter" .. abs(arrange)]:SetBackdropColor(0.2, 0.2, 0.2)
						else
							bg.name:SetBackdropColor(0.2, 0.2, 0.2)
						end
						
						arrange = a
					end
					
					self:ClearFocus()
					sorting = true
					tasks[#tasks + 1] = create(Sort)
				end
			end
		end)
		
		bg[filter]:SetScript("OnEditFocusGained", function(self)
			if self:GetText() == filterLabels[a] then
				self:SetText("")
				self:SetTextColor(0.95, 0.95, 0.95)
				self:SetNumeric(true)
			end
			
			if self:GetTextInsets() == 6 then
				self:ClearFocus()
			end
		end)
		
		bg[filter]:SetScript("OnEditFocusLost", function(self)
			self:HighlightText(0, 0)
			
			if self:GetText() == "" then
				self:SetNumeric(false)
				self:SetTextColor(0.961, 0.871, 0.702)
				self:SetText(filterLabels[a])
			end
		end)
		
		bg[filter]:SetScript("OnEnterPressed", function(self)
			self:ClearFocus()
			bg.search:Click()
		end)
		
		bg[filter]:SetScript("OnEnter", function(self)
			if abs(arrange) ~= a then
				self:SetBackdropColor(0.15, 0.15, 0.15)
			end
		end)
		
		bg[filter]:SetScript("OnLeave", function(self)
			if abs(arrange) ~= a then
				self:SetBackdropColor(0.2, 0.2, 0.2)
			end
		end)
	end
	
	bg.filter3:SetScript("OnEditFocusLost", function(self)
		self:HighlightText(0, 0)
		
		if self:GetText() == "" then
			self:SetNumeric(false)
			self:SetTextColor(0.961, 0.871, 0.702)
			
			if bg.weights:GetChecked() then
				self:SetText("£")
			else
				self:SetText("D/Ar")
			end
		end
	end)
	
	bg.weights = CreateFrame("CheckButton", nil, bg, "BackdropTemplate")
	bg.weights:SetPoint("TOPRIGHT", -14, -42)
	bg.weights:SetSize(24, 24)
	bg.weights.text = bg.weights:CreateFontString()
	bg.weights.text:SetPoint("CENTER", 0, -1)
	bg.weights.text:SetText("£")
	bg.weights.texture = bg.weights:CreateTexture()
	bg.weights.texture:SetColorTexture(0.3, 0.3, 0.3)
	bg.weights.texture:SetPoint("CENTER")
	bg.weights.texture:SetSize(18, 18)
	bg.weights:SetCheckedTexture(bg.weights.texture)
	bg.weights:SetScript("OnMouseDown", function(self)
		self.text:SetPoint("CENTER", 2, -3)
	end)
	
	bg.weights:SetScript("OnMouseUp", function(self)
		self.text:SetPoint("CENTER", 0, -1)
	end)
	
	-- Results
	local result
	local stat
	
	for a = 1, 20 do
		result = "result" .. a
		
		bg[result] = CreateFrame("Frame", nil, bg, "BackdropTemplate")
		bg[result]:SetPoint("TOPLEFT", 14, -42 - 28 * a)
		bg[result]:SetSize(992, 24)
		bg[result]:SetHitRectInsets(0, 0, -2, -2)
		bg[result]:Hide()
		bg[result].icon = bg[result]:CreateTexture()
		bg[result].icon:SetPoint("LEFT", 4, 0)
		bg[result].icon:SetSize(20, 20)
		bg[result].icon:SetTexCoord(0.1, 0.9, 0.1, 0.9)
		bg[result].link = bg[result]:CreateFontString()
		bg[result].link:SetPoint("LEFT", 28, 0)
		bg[result].link:SetWidth(316)
		bg[result].link:SetJustifyH("LEFT")
		bg[result].link:SetWordWrap(false)
		
		for b = 1, 12 do
			stat = "stat" .. b
			
			bg[result][stat] = bg[result]:CreateFontString()
			bg[result][stat]:SetPoint("LEFT", 294 + b * 54, 0)
		end
		
		local offset
		local delta
		local itemString
		
		bg[result]:SetScript("OnEnter", function(self)
			self:SetBackdropColor(0.15, 0.15, 0.15)
			GameTooltip:SetOwner(bg, "ANCHOR_BOTTOMLEFT", 0, 642)
			offset = results[(page + a - 2) * 10 + 1]
			itemString = {"item:", items[offset + 1], "::::::::::::", 0}
			
			-- Item level scaling
			if filters[14] then
				delta = filters[14] - items[offset + 7]
				
				if delta > 0 and up then
					itemString[4] = floor(abs(delta * 0.005) + 1.0005)
					
					for c = 1, itemString[4] - 1 do
						itemString[c + 4] = ":1672"
					end
					
					itemString[#itemString + 1] = ":"
					itemString[#itemString + 1] = 1472 + delta - 200 * (itemString[4] - 1)
				elseif delta < 0 and down then
					itemString[4] = floor(abs(delta * 0.01) + 1.005)
					
					for c = 1, itemString[4] - 1 do
						itemString[c + 4] = ":1372"
					end
					
					itemString[#itemString + 1] = ":"
					itemString[#itemString + 1] = 1472 + delta + 100 * (itemString[4] - 1)
				end
			end
			
			-- Azerite
			if C_AzeriteEmpoweredItem.IsAzeriteEmpoweredItemByID(itemString[2]) and
					not C_AzeriteEmpoweredItem.IsAzeritePreviewSourceDisplayable(itemString[2])	then
				itemString[4] = itemString[4] + 1
				itemString[#itemString + 1] = ":"
				offset = #itemString + 1
				
				for a = 6526, 3569, -1 do -- bruh
					itemString[offset] = a
					
					if C_AzeriteEmpoweredItem.IsAzeritePreviewSourceDisplayable(concat(itemString)) then
						break
					end
				end
			end
			
			GameTooltip:SetHyperlink(concat(itemString))
			GameTooltip:Show()
		end)
		
		bg[result]:SetScript("OnLeave", function(self)
			self:SetBackdropColor(0.2, 0.2, 0.2)
			GameTooltip:Hide()
		end)
		
		bg[result]:SetScript("OnMouseUp", function(_, button)
			local itemLink = ({GameTooltip:GetItem()})[2]
			
			if itemLink then
				SetItemRef(itemLink, itemLink, button)
			end
		end)
	end
	
	-- Scrolling
	bg.scroll = CreateFrame("Slider", nil, bg, "OptionsSliderTemplate")
	bg.scroll:SetPoint("BOTTOMRIGHT", -14, 42)
	bg.scroll:SetSize(24, 500)
	bg.scroll:SetMinMaxValues(1, 1)
	bg.scroll:SetValue(1)
	bg.scroll:SetValueStep(1)
	bg.scroll:SetObeyStepOnDrag(true)
	bg.scroll:SetOrientation("VERTICAL")
	bg.scroll:SetScript("OnValueChanged", function(_, value)
		if value ~= page then
			page = value
			Display()
		end
	end)
	
	bg.up = CreateFrame("Button", nil, bg, "BackdropTemplate")
	bg.up:SetPoint("TOPRIGHT", -14, -70)
	bg.up:SetSize(24, 24)
	bg.up:SetFrameLevel(12)
	bg.up.text = bg.up:CreateFontString()
	bg.up.text:SetPoint("CENTER", 0, -2)
	bg.up.text:SetFont("Interface\\Addons\\Chardev\\Assets\\Fonts\\Nunito\\Nunito-Regular.ttf", 22, "")
	bg.up.text:SetText("^")
	bg.up:SetScript("OnMouseDown", function(self) self.text:SetPoint("CENTER", 2, -4) end)
	bg.up:SetScript("OnMouseUp", function(self, button)
		self.text:SetPoint("CENTER", 0, -2)
		
		if button == "LeftButton" and self:IsMouseOver() then
			bg.scroll:SetValue(page - 1)
		end
	end)
	
	bg.down = CreateFrame("Button", nil, bg, "BackdropTemplate")
	bg.down:SetPoint("BOTTOMRIGHT", -14, 14)
	bg.down:SetSize(24, 24)
	bg.down:SetFrameLevel(12)
	bg.down.text = bg.down:CreateFontString()
	bg.down.text:SetPoint("CENTER")
	bg.down.text:SetText("V")
	bg.down:SetScript("OnMouseDown", function(self) self.text:SetPoint("CENTER", 2, -2) end)
	bg.down:SetScript("OnMouseUp", function(self, button)
		self.text:SetPoint("CENTER")
		
		if button == "LeftButton" and self:IsMouseOver() then
			bg.scroll:SetValue(page + 1)
		end
	end)
end

SlashCmdList["Chardev"] = Chardev_Slash

SLASH_Chardev1 = "/chardev"
SLASH_Chardev2 = "/cd"
